import abc
import glob
import json
import os
import pickle
import re
from functools import cached_property
from dataclasses import dataclass, fields

from searchkit.utils import MPCache
from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers.exceptions import (
    CLIExecError,
    SourceNotFound,
)
from hotsos.core.log import log
from hotsos.core.utils import sorted_dict


class NullCache():
    """ A cache that does nothing but maintains the MPCache abi. """

    @staticmethod
    def get(*args, **kwargs):
        log.debug("null cache get() op args=%s kwargs=%s", args, kwargs)

    @staticmethod
    def set(*args, **kwargs):
        log.debug("null cache set() op args=%s kwargs=%s", args, kwargs)


class HostHelpersBase(abc.ABC):
    """ Base class for all hosthelpers. """
    def __init__(self, *args, **kwargs):
        if not self.cache_root or not os.path.exists(self.cache_root):
            log.debug("cache root invalid or does not exist so disabling %s "
                      "cache", self.__class__.__name__)
            self.cache = NullCache()
        else:
            self.cache = MPCache(self.cache_name,
                                 f'host_helpers_{self.cache_type}',
                                 self.cache_root)

        super().__init__(*args, **kwargs)

    @property
    def cache_root(self):
        """
        By default caches are global to all plugins but this can be overridden
        if we want otherwise.
        """
        return HotSOSConfig.global_tmp_dir

    @property
    @abc.abstractmethod
    def cache_name(self):
        """ Unique name for cache used by instance of this object. """

    @property
    @abc.abstractmethod
    def cache_type(self):
        """
        Unique name for the type of cache used by instance of this object.
        """

    @abc.abstractmethod
    def cache_load(self, key):
        """ Load cache contents. """

    @abc.abstractmethod
    def cache_save(self, key, value):
        """ Save contents to cache. """


class ServiceManagerBase(abc.ABC):
    """ Base class for service manager helper implementations. """
    PS_CMD_EXPR_TEMPLATES = {
        'absolute': r".+\S+bin/({})(?:\s+.+|$)",
        'snap': r".+\S+\d+/({})(?:\s+.+|$)",
        'relative': r".+\s({})(?:\s+.+|$)",
    }

    def __init__(self, service_exprs, ps_allow_relative=True):
        """
        @param service_exprs: list of python.re expressions used to match
                              service names.
        @param ps_allow_relative: whether to allow commands to be identified
                                  from ps as run using an relative binary
                                  path e.g. mycmd as opposed to /bin/mycmd.
        """
        self._ps_allow_relative = ps_allow_relative
        self._service_exprs = set(service_exprs)

    @property
    @abc.abstractmethod
    def _service_manager_type(self):
        """ A string name representing the type of service manager e.g.
        'systemd' """

    def get_cmd_from_ps_line(self, line, expr):
        """
        Match a command in ps output line.

        @param line: line from ps output
        @param expr: regex to match a command. See PS_CMD_EXPR_TEMPLATES.
        @param return: matched command name.
        """
        for expr_type, expr_tmplt in self.PS_CMD_EXPR_TEMPLATES.items():
            if expr_type == 'relative' and not self._ps_allow_relative:
                continue

            ret = re.compile(expr_tmplt.format(expr)).match(line)
            if ret:
                cmd = ret.group(1)
                log.debug("matched command '%s' with expr type '%s'", cmd,
                          expr_type)
                return cmd

        return None

    @property
    @abc.abstractmethod
    def _service_filtered_ps(self):
        """ Return a list ps entries corresponding to services. """

    @cached_property
    def processes(self):
        """
        Identify running processes from ps that are associated with resolved
        services. The search pattern used to identify a service is also
        used to match the process binaryc/cmd name.

        Accounts for different types of process cmd path e.g.

        /snap/<name>/1830/<svc>
        /usr/bin/<svc>

        and filter e.g.

        /var/lib/<svc> and /var/log/<svc>

        Returns a dictionary of process names along with the number of each.
        """
        _proc_info = {}
        for line in self._service_filtered_ps:
            for expr in self._service_exprs:
                cmd = self.get_cmd_from_ps_line(line, expr)
                if not cmd:
                    continue

                if cmd in _proc_info:
                    _proc_info[cmd] += 1
                else:
                    _proc_info[cmd] = 1

        return _proc_info

    @property
    @abc.abstractmethod
    def services(self):
        """ Return a dictionary of identified services and their state. """

    @property
    def _service_info(self):
        """Return a dictionary of services grouped by state. """
        info = {}
        for svc, obj in sorted_dict(self.services).items():
            state = obj.state
            if state not in info:
                info[state] = []

            info[state].append(svc)

        return info

    @property
    def _process_info(self):
        """Return a list of processes associated with services. """
        return [f"{name} ({count})"
                for name, count in sorted_dict(self.processes).items()]

    @property
    def summary(self):
        """
        Output a dict summary of this class i.e. services, their state and any
        processes run by them.
        """
        return {self._service_manager_type: self._service_info,
                'ps': self._process_info}


class NullSource():
    """ Exception raised to indicate that a datasource is not available. """
    def __call__(self, *args, **kwargs):
        return CmdOutput([])


@dataclass(frozen=True)
class CmdOutput():
    """ Representation of the output of a command. """

    # Output value.
    value: str
    # Optional command source path.
    source: str = None


class SourceRunner():
    """ Manager to control how we execute commands.

    Ensures that we try data sources in a consistent order.
    """
    def __init__(self, cmdkey, sources, cache, output_file=None):
        """
        @param cmdkey: unique key identifying this command.
        @param sources: list of command source implementations.
        @param cache: CLICacheWrapper object.
        """
        self.cmdkey = cmdkey
        self.sources = sources
        self.cache = cache
        self.output_file = output_file
        # Command output can differ between CLIHelper and CLIHelperFile so we
        # need to cache them separately.
        if output_file:
            self.cache_cmdkey = f"{cmdkey}.file"
        else:
            self.cache_cmdkey = cmdkey

    def bsource(self, *args, **kwargs):
        # binary sources only apply if data_root is system root
        bin_out = None
        for bsource in [s for s in self.sources if s.TYPE == "BIN"]:
            cache = False
            # NOTE: we currently only support caching commands with no
            #       args.
            if not any([args, kwargs]):
                cache = True
                out = self.cache.load(self.cache_cmdkey)
                if out is not None:
                    return out

            try:
                if self.output_file:
                    # don't decode if we are going to be saving to a file
                    kwargs['skip_json_decode'] = True

                bin_out = bsource(*args, **kwargs)
                if cache and bin_out is not None:
                    try:
                        self.cache.save(self.cache_cmdkey, bin_out)
                    except pickle.PicklingError as exc:
                        log.info("unable to cache command '%s' output: %s",
                                 self.cmdkey, exc)

                # if command executed but returned nothing that still counts
                # as success.
                break
            except CLIExecError as exc:
                bin_out = CmdOutput(exc.return_value)

        return bin_out

    def fsource(self, *args, **kwargs):
        for fsource in [s for s in self.sources if s.TYPE == "FILE"]:
            try:
                skip_load_contents = False
                if self.output_file:
                    skip_load_contents = True

                return fsource(*args, **kwargs,
                               skip_load_contents=skip_load_contents)
            except CLIExecError as exc:
                return CmdOutput(exc.return_value)
            except SourceNotFound:
                pass

        return None

    def _execute(self, *args, **kwargs):
        # always try file sources first
        ret = self.fsource(*args, **kwargs)
        if ret is not None:
            return ret

        if HotSOSConfig.data_root != '/':
            return NullSource()()

        return self.bsource(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        """
        Execute the command using the appropriate source runner. These can be
        binary or file-based depending on whether data root points to / or a
        sosreport. File-based are attempted first.

        A command can have more than one source implementation so we must
        ensure they all have a chance to run.
        """
        out = self._execute(*args, **kwargs)
        if self.output_file:
            if out.source is not None:
                return out.source

            with open(self.output_file, 'w', encoding='utf-8') as fd:
                if isinstance(out.value, list):
                    fd.write(''.join(out.value))
                elif isinstance(out.value, dict):
                    fd.write(json.dumps(out.value))
                else:
                    fd.write(out.value)

                return self.output_file

        return out.value


def run_pre_exec_hooks(f):
    """ pre-exec hooks are run before running __call__ method.

    These hooks are not expected to return anything and are used to manipulate
    the instance variables used by the main __call__ method.
    """
    def run_pre_exec_hooks_inner(self, *args, **kwargs):
        hook = self.hooks.get("pre-exec")
        if hook:
            # no return expected
            hook(*args, **kwargs)

        return f(self, *args, **kwargs)

    return run_pre_exec_hooks_inner


def run_post_exec_hooks(f):
    """ post-exec hooks are run after running __call__ method and take its
    output as input.
    """
    def run_post_exec_hooks_inner(self, *args, **kwargs):
        out = f(self, *args, **kwargs)
        hook = self.hooks.get("post-exec")
        if hook:
            out = hook(out, *args, **kwargs)

        return out

    return run_post_exec_hooks_inner


def reset_command(f):
    """
    This should be run by all commands as their last action after all/any hooks
    have run.
    """
    def reset_command_inner(self, *args, **kwargs):
        out = f(self, *args, **kwargs)
        self.reset()
        return out

    return reset_command_inner


def get_ps_axo_flags_available():
    path = os.path.join(HotSOSConfig.data_root,
                        "sos_commands/process/ps_axo_flags_state_"
                        "uid_pid_ppid_pgid_sid_cls_pri_addr_sz_wchan*_lstart_"
                        "tty_time_cmd")
    _paths = []
    for path in glob.glob(path):
        _paths.append(path)

    if not _paths:
        return None

    # strip data_root since it will be prepended later
    return _paths[0].partition(HotSOSConfig.data_root)[2]


@dataclass
class InstallInfoBase():
    """
    Provides a common way to define install information such as packages and
    runtime services assocated with a plugin.

    To use this class it should be re-implemented as a dataclass that
    initialises the required attributes. It can then either be inherited or
    mixed in to avoid issues with excessive inheritance.
    """
    apt: None = None
    pebble: None = None
    snaps: None = None
    systemd: None = None

    def mixin(self, _self):
        for attr in fields(self):
            val = getattr(self, attr.name)
            if val is None:
                continue

            setattr(_self, attr.name, val)



import abc


# pylint: disable-next= too-few-public-methods
class FactoryBase(abc.ABC):
    """
    Provide a common way to implement factory objects.

    The basic idea is that implementations of this class are instantiated and
    then content is generated using attrs as input. This provides a way e.g. to
    defer operations on a set of data that need only be retrieved once.
    """

    @abc.abstractmethod
    def __getattr__(self, name):
        """
        All factory implementations must implement this method to
        allow them to dynamically generate objects from arbitrary input
        provided by calling as an attribute on this object.
        """



import re
from dataclasses import fields

from hotsos.core.host_helpers import CLIHelper, HostNetworkingHelper
from hotsos.core.plugins.openstack.common import OpenstackBase, OpenStackChecks
from hotsos.core.plugins.openstack.neutron import (
    IP_HEADER_BYTES,
    GRE_HEADER_BYTES,
    VXLAN_HEADER_BYTES,
)
from hotsos.core.plugins.openvswitch import OpenvSwitchBase
from hotsos.core.issues import (
    IssuesManager,
    OpenstackWarning,
)
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)


class OpenstackNetworkChecks(OpenstackBase, OpenStackChecks):
    """ Implements OpenStack network checks. """
    summary_part_index = 4

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cli = CLIHelper()

    @property
    def summary_subkey(self):
        return 'network'

    @staticmethod
    def _get_port_stat_outliers(counters):
        """ For a given port's packet counters, identify outliers i.e. > 1%
        and create a new dict with count and percent values.
        """
        stats = {}
        for rxtx in counters:
            total = sum(counters[rxtx].values())
            for key, value in counters[rxtx].items():
                if key == "packets":
                    continue

                if value:
                    pcent = int(100 / float(total) * float(value))
                    if pcent <= 1:
                        continue

                    if rxtx not in stats:
                        stats[rxtx] = {}

                    stats[rxtx][key] = f"{int(value)} ({pcent}%)"

        return stats

    def get_config_info(self):
        config_info = {}
        for project in [f.name for f in fields(self.project_helpers)]:
            _project = getattr(self.project_helpers, project)
            if _project.bind_interfaces:
                for name, port in _project.bind_interfaces.items():
                    if project not in config_info:
                        config_info[project] = {}

                    config_info[project][name] = port.to_dict()

        return config_info

    def get_phy_port_health_info(self):
        """ Identify ports used by Openstack services, include them in output
        for informational purposes along with their health (dropped packets
        etc) for any outliers detected.
        """
        port_health_info = {}
        for project in [f.name for f in fields(self.project_helpers)]:
            _project = getattr(self.project_helpers, project)
            if not (_project and _project.bind_interfaces):
                continue

            for port in _project.bind_interfaces.values():
                if port.stats:
                    stats = self._get_port_stat_outliers(port.stats)
                    if not stats:
                        continue

                    port_health_info[port.name] = stats

        return port_health_info

    @summary_entry('config', get_min_available_entry_index() + 5)
    def summary_config(self):
        config_info = self.get_config_info()
        if config_info:
            return config_info

        return None

    @summary_entry('phy-port-health', get_min_available_entry_index() + 6)
    def summary_phy_port_health(self):
        port_health_info = self.get_phy_port_health_info()
        if port_health_info:
            return port_health_info

        return None

    @summary_entry('namespaces', get_min_available_entry_index() + 7)
    def summary_namespaces(self):
        """Populate namespace information dict."""
        ns_info = {}
        for line in self.cli.ip_netns():
            ret = re.compile(r"^([a-z0-9]+)-([0-9a-z\-]+)\s+.+").match(line)
            if ret:
                if ret[1] in ns_info:
                    ns_info[ret[1]] += 1
                else:
                    ns_info[ret[1]] = 1

        if ns_info:
            return ns_info

        return None

    def _get_router_iface_mtus(self):
        """
        Get the mtu of each qr and qg port in every qrouter or snat namespace.

        @return: dictionary of port prefixes and list of mtus associated with
                 those prefixes.
        """
        router_mtus = {}
        for ns in self.cli.ip_netns():
            # strip index
            ns = ns.partition(" ")[0]
            for nsprefix in ['qrouter', 'snat']:
                if not ns.startswith(nsprefix):
                    continue

                for port in HostNetworkingHelper().get_ns_interfaces(ns):
                    for pprefix in ['qr', 'sg']:
                        if not port.name.startswith(pprefix):
                            continue

                        if pprefix not in router_mtus:
                            router_mtus[pprefix] = set()

                        router_mtus[pprefix].add(port.mtu)

        return {prefix: list(mtus) for prefix, mtus in router_mtus.items()}

    @summary_entry('router-port-mtus', get_min_available_entry_index() + 8)
    def summary_router_port_mtus(self):
        """ Provide a summary of ml2-ovs router port mtus. """
        bind_interfaces = self.project_helpers.neutron.bind_interfaces
        if not bind_interfaces:
            return None

        phy_mtus = set()
        for port in bind_interfaces.values():
            phy_mtus.add(port.mtu)

        tunnels = OpenvSwitchBase().tunnels
        # ip header
        overhead = IP_HEADER_BYTES
        if 'vxlan' in tunnels:
            overhead += VXLAN_HEADER_BYTES
        else:
            # gre
            overhead += GRE_HEADER_BYTES

        router_mtus = self._get_router_iface_mtus()
        all_router_mtus = set()
        for mtus in router_mtus.values():
            all_router_mtus.update(mtus)

        if phy_mtus and all_router_mtus:
            smallest_allowed = min(phy_mtus) - overhead
            if max(all_router_mtus) > smallest_allowed:
                msg = ("This Neutron L3 agent host has one or more router "
                       f"ports with mtu={max(all_router_mtus)} which is "
                       "greater or equal to the smallest allowed "
                       f"({smallest_allowed}) on the physical network. "
                       "This will result in dropped packets or "
                       "unexpected fragmentation in overlay networks.")
                IssuesManager().add(OpenstackWarning(msg))

        return router_mtus

    @summary_entry('vm-port-health', get_min_available_entry_index() + 9)
    def summary_vm_port_health(self):
        """ For each instance get its ports and check port health, reporting on
        any outliers. """
        if not self.project_helpers.nova.instances:
            return None

        port_health_info = {}
        for guest in self.project_helpers.nova.instances.values():
            for port in guest.ports:
                stats = port.stats
                if stats:
                    outliers = self._get_port_stat_outliers(stats)
                    if not outliers:
                        continue

                    if guest.uuid not in port_health_info:
                        port_health_info[guest.uuid] = {}

                    port_health_info[guest.uuid][port.hwaddr] = outliers

        if port_health_info:
            health = {'num-vms-checked':
                      len(self.project_helpers.nova.instances),
                      'stats': port_health_info}
            return health

        return None



from dataclasses import dataclass, field
from functools import cached_property

from hotsos.core.host_helpers import (
    APTPackageHelper,
    CLIHelperFile,
    InstallInfoBase,
    SnapPackageHelper,
    SystemdHelper,
)
from hotsos.core.plugintools import PluginPartBase
from hotsos.core.search import (
    FileSearcher, SearchDef,
    SequenceSearchDef
)


CORE_APT = ['lxd', 'lxc']
CORE_SNAPS = [rf"(?:snap\.)?{p}" for p in CORE_APT]
SERVICE_EXPRS = [rf"{s}\S*" for s in CORE_SNAPS]


class LXD():
    """ LXD interface. """
    @cached_property
    def instances(self):
        """ Return a list of instance names. """
        _instances = []
        s = FileSearcher()
        seq = SequenceSearchDef(start=SearchDef(r'^## Instances$'),
                                body=SearchDef(r'^\|\s+(\S+)\s+\|'),
                                end=SearchDef(r'##.*'),
                                tag='instances')
        with CLIHelperFile() as cli:
            s.add(seq, path=cli.lxd_buginfo())
            results = s.run()
            for section in results.find_sequence_sections(seq).values():
                for r in section:
                    if 'body' in r.tag:
                        if r.get(1) != 'NAME' and r.get(1) != '|':
                            _instances.append(r.get(1))

        return _instances


@dataclass
class LXDInstallInfo(InstallInfoBase):
    """ LXD installation information. """
    apt: APTPackageHelper = field(default_factory=lambda:
                                  APTPackageHelper(core_pkgs=CORE_APT))
    snaps: SnapPackageHelper = field(default_factory=lambda:
                                     SnapPackageHelper(core_snaps=CORE_SNAPS))
    systemd: SystemdHelper = field(default_factory=lambda:
                                   SystemdHelper(service_exprs=SERVICE_EXPRS))


class LXDChecks(PluginPartBase):
    """ LXD Checks. """
    plugin_name = 'lxd'
    plugin_root_index = 11

    def __init__(self, *args, **kwargs):
        super().__init__()
        LXDInstallInfo().mixin(self)

    @classmethod
    def is_runnable(cls):
        """
        Determine whether or not this plugin can and should be run.

        @return: True or False
        """
        lxd_pkgs = LXDInstallInfo()
        if lxd_pkgs.apt.core or lxd_pkgs.snaps.core:
            return True

        return False



from hotsos.core.log import log
from hotsos.core.plugins.openstack.common import OpenstackBase, OpenStackChecks
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)

FEATURES = {'neutron': {
                'main': {
                    'DEFAULT': ['global_physnet_mtu'],
                    'AGENT': ['availability_zone']},
                'openvswitch-agent': {
                    'AGENT': ['l2_population'],
                    'SECURITYGROUP': ['firewall_driver']},
                'l3-agent': {
                    'DEFAULT': ['agent_mode', 'ovs_use_veth']},
                'dhcp-agent': {
                    'DEFAULT': ['enable_metadata_network',
                                'enable_isolated_metadata',
                                'ovs_use_veth']},
                'ovn': {
                    'DEFAULT': ['enable_distributed_floating_ip']},
                'ml2': {
                    'ML2': ['path_mtu']}},
            'nova': {
                'main': {
                    'DEFAULT': ['vcpu_pin_set'],
                    'compute': [
                                'cpu_shared_set',
                                'cpu_dedicated_set'],
                    'libvirt': ['cpu_mode',
                                'cpu_model',
                                'cpu_model_extra_flags',
                                'live_migration_permit_auto_converge',
                                'live_migration_permit_post_copy']}}}

# checked against neutron
DEFAULTS = {'neutron': {'dhcp-agent': {
                            'enable_metadata_network': False,
                            'enable_isolated_metadata': False}},
            'nova': {'main': {'live_migration_permit_auto_converge': False,
                              'live_migration_permit_post_copy': False}}}


class ServiceFeatureChecks(OpenstackBase, OpenStackChecks):
    """ Implements Openstack service feature checks. """
    summary_part_index = 5

    @staticmethod
    def _get_module_features(service, module, sections, cfg):
        """
        Get dictionary of features for this service module and whether they are
        enabled/disabled.

        @param service: name of Openstack service
        @param module: service submodule (agent)
        @param sections: config sections
        @param cfg: OpenstackConfig object
        """
        module_features = {}
        for section, keys in sections.items():
            for key in keys:
                val = cfg.get(key, section=section)
                if val is not None:
                    module_features[key] = val

                if key in module_features or service not in DEFAULTS:
                    continue

                if (module not in DEFAULTS[service] or
                        key not in DEFAULTS[service][module]):
                    continue

                module_features[key] = DEFAULTS[service][module][key]

        return module_features

    @summary_entry('features', get_min_available_entry_index() + 3)
    def summary_features(self):
        """
        This is used to display whether or not specific features are enabled.
        """
        features = {}
        for service, modules in FEATURES.items():
            svc_cfg = self.project_catalog.all[service].config
            if not svc_cfg['main'].exists:
                continue

            dbg_enabled = svc_cfg['main'].get('debug', section="DEFAULT")
            features[service] = {'main': {'debug': dbg_enabled or False}}

            for module, sections in modules.items():
                log.debug("getting features for '%s.%s'", service, module)
                cfg = svc_cfg[module]
                if not cfg.exists:
                    log.debug("%s not found - skipping features", cfg.path)
                    continue

                module_features = self._get_module_features(service, module,
                                                            sections, cfg)
                if not module_features:
                    continue

                if module in features[service]:
                    features[service][module].update(module_features)
                else:
                    features[service][module] = module_features

        if self.ssl_enabled is not None:
            features['api-ssl'] = self.ssl_enabled or False

        return features or None



import os
from datetime import timedelta
from functools import cached_property

from searchkit import (
    FileSearcher as _FileSearcher,
    ResultFieldInfo,
    SearchDef,
    SequenceSearchDef,
)
from searchkit.constraints import (
    TimestampMatcherBase,
    SearchConstraintSearchSince as _SearchConstraintSearchSince
)
from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers import CLIHelper, UptimeHelper
from hotsos.core.log import log


# This module acts as a proxy to searchkit but with some addons/modifications
__all__ = [
    ResultFieldInfo.__name__,
    SearchDef.__name__,
    SequenceSearchDef.__name__,
    ]


class FileSearcher(_FileSearcher):
    """
    Custom representation of searchkit FilesSearcher that sets some parameters
    based on HotSOSConfig that we want to apply to all use cases.
    """
    def __init__(self, *args, **kwargs):
        if HotSOSConfig.use_all_logs:
            max_logrotate_depth = HotSOSConfig.max_logrotate_depth
        else:
            max_logrotate_depth = 1

        super().__init__(*args,
                         max_parallel_tasks=HotSOSConfig.max_parallel_tasks,
                         max_logrotate_depth=max_logrotate_depth,
                         **kwargs)


class SearchConstraintSearchSince(_SearchConstraintSearchSince):
    """
    Custom representation of searchkit SearchConstraintSearchSince that
    automatically applies global log history limiting.
    """
    def __init__(self, *args, **kwargs):
        if 'days' not in kwargs and 'hours' not in kwargs:
            days = 1
            if HotSOSConfig.use_all_logs:
                days = HotSOSConfig.max_logrotate_depth

            kwargs['days'] = days

        current_date = CLIHelper().date(format='+%Y-%m-%d %H:%M:%S')
        if not current_date or not isinstance(current_date, str):
            log.error("current date '%s' being provided to search "
                      "constraints is not valid.", current_date)
            return

        super().__init__(*args, current_date=current_date, **kwargs)

    def apply_to_line(self, *args, **kwargs):
        if not os.path.isdir(os.path.join(HotSOSConfig.data_root,
                                          'sos_commands')):
            log.info("skipping line constraint since data_root is not a "
                     "sosreport therefore files may be changing")
            return True

        return super().apply_to_line(*args, **kwargs)

    def apply_to_file(self, *args, **kwargs):
        if not os.path.isdir(os.path.join(HotSOSConfig.data_root,
                                          'sos_commands')):
            log.info("skipping file constraint since data_root is not a "
                     "sosreport therefore files may be changing")
            return 0

        return super().apply_to_file(*args, **kwargs)


class CommonTimestampMatcher(TimestampMatcherBase):
    """
    This class must support regex patterns to match any kind of timestamp that
    we would expect to find. When a plugin results in the search of file
    (typically log files) that contain timestamps it is necessary to ensure
    the patterns in this class support matching those timestamps in order for
    search constraints to work.

    TODO: timestamps typically use an RFC format so we should brand them
          as such here.
    """
    MONTH_MAP = {'jan': 1, 'feb': 2, 'mar': 3, 'apr': 4, 'may': 5,
                 'jun': 6, 'jul': 7, 'aug': 8, 'sep': 9, 'oct': 10,
                 'nov': 11, 'dec': 12}

    @cached_property
    def _current_year(self):
        return CLIHelper().date(format='+%Y')

    @property
    def year(self):
        """ Needed for kernlog which has no year group. """
        try:
            return self.result.group('year')
        except IndexError:
            pass

        return self._current_year

    @property
    def month(self):
        """ Needed for kernlog which has a string month. """
        try:
            return int(self.result.group('month'))
        except ValueError:
            pass

        _month = self.result.group('month').lower()
        try:
            return self.MONTH_MAP[_month[:3]]
        except KeyError:
            log.exception("could not establish month integer from '%s'",
                          _month)

        return None

    @property
    def patterns(self):
        """
        This needs to contain timestamp patterns for any/all types of file
        we want to analyse where SearchConstraintsSince is to be applied.
        """
        # should match plugins.openstack.openstack.OpenstackDateTimeMatcher
        openstack = (r'^(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})+\s+'
                     r'(?P<hours>\d{2}):(?P<minutes>\d{2}):(?P<seconds>\d+)')
        # since they are identical we wont add but leaving in case we want to.
        # edit later.
        # juju = openstack
        # should match plugins.storage.ceph.CephDateTimeMatcher
        ceph = (r'^(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})[\sT]'
                r'(?P<hours>\d{2}):(?P<minutes>\d{2}):(?P<seconds>\d+)')
        # since they are identical we wont add but leaving in case we want to
        # edit later.
        # openvswitch = ceph
        kernlog = (r'^(?P<month>\w{3,5})\s+(?P<day>\d{1,2})\s+'
                   r'(?P<hours>\d{2}):(?P<minutes>\d{2}):(?P<seconds>\d{2})')
        return [openstack, ceph, kernlog]


class ExtraSearchConstraints():
    """
    Provides a way to apply constraints to search results that are not yet
    natively supported by searchkit.
    """

    @classmethod
    def _get_datetime_from_result(cls, result):
        """
        This attempts to create a datetime object from a timestamp (usually
        from a log file) extracted from a search result. If it is not able
        to do so it will return None. The normal expectation is that two search
        result groups be available at index 1 and 2 but if only 1 is valid it
        will be used a fallback.
        """
        ts = result.get(1)
        ts = f"{ts} {result.get(2) or '00:00:00'}"

        ts_matcher = CommonTimestampMatcher(ts)
        if not ts_matcher.matched:
            log.warning("failed to parse timestamp string '%s' (num_group=%s) "
                        "- returning None", ts, len(result))
            return None

        return ts_matcher.strptime

    @classmethod
    def filter_by_period(cls, results, period_hours):
        """ Return the most recent period_hours worth of results. """
        if not period_hours:
            log.debug("period filter not specified - skipping")
            return results

        log.debug("applying search filter (period_hours=%s)", period_hours)

        _results = []
        for r in results:
            ts = cls._get_datetime_from_result(r)
            if ts:
                _results.append((ts, r))

        results = []
        last = None

        for r in sorted(_results, key=lambda i: i[0], reverse=True):
            if last is None:
                last = r[0]
            elif r[0] < last - timedelta(hours=period_hours):
                break

            results.append(r)

        log.debug("%s results remain after applying filter", len(results))
        return [r[1] for r in results]

    def apply(self, results, search_period_hours=None, min_results=None):
        if results:
            results = self.filter_by_period(results, search_period_hours)

        if min_results is None:
            return results

        count = len(results)
        if count < min_results:
            log.debug("search does not have enough matches (%s) to "
                      "satisfy min of %s", count, min_results)
            return []

        log.debug("applying extra search constraints reduced results from %s "
                  "to %s", count, len(results))
        return results


def create_constraint(search_result_age_hours=None,
                      min_hours_since_last_boot=None):
    """
    Create a SearchConstraintSearchSince object if necessary (if one of
    search_result_age_hours or min_hours_since_last_boot is not None).
    """
    if not any([search_result_age_hours, min_hours_since_last_boot]):
        return None

    uptime_etime_hours = UptimeHelper().in_hours
    hours = search_result_age_hours
    if not hours:
        hours = max(uptime_etime_hours - min_hours_since_last_boot, 0)
    elif min_hours_since_last_boot > 0:
        hours = min(hours,
                    max(uptime_etime_hours - min_hours_since_last_boot, 0))

    return SearchConstraintSearchSince(
                                     ts_matcher_cls=CommonTimestampMatcher,
                                     hours=hours)



from .common import (
    OpenstackBase,
    OpenStackChecks,
    OpenstackEventHandlerBase,
    OpenstackEventCallbackBase,
)
from .openstack import OpenstackConfig

__all__ = [
    OpenstackBase.__name__,
    OpenStackChecks.__name__,
    OpenstackConfig.__name__,
    OpenstackEventHandlerBase.__name__,
    OpenstackEventCallbackBase.__name__,
    ]



# flake8: noqa
# pylint: disable=C0301

# From:
#  - https://github.com/openstack/octavia/blob/master/octavia/common/exceptions.py
#  - https://github.com/openstack/octavia/blob/master/octavia/amphorae/driver_exceptions/exceptions.py
# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' octavia/common/exceptions.py
# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' octavia/network/base.py
OCTAVIA_EXCEPTIONS = [
# common/exceptions.py
    "OctaviaException",
    "APIException",
    "NotFound",
    "PolicyForbidden",
    "InvalidOption",
    "InvalidFilterArgument",
    "DisabledOption",
    "L7RuleValidation",
    "SingleCreateDetailsMissing",
    "InvalidHMACException",
    "MissingArguments",
    "NetworkConfig",
    "NeedsPassphrase",
    "UnreadableCert",
    "UnreadablePKCS12",
    "MisMatchedKey",
    "CertificateRetrievalException",
    "CertificateStorageException",
    "CertificateGenerationException",
    "DuplicateListenerEntry",
    "DuplicateMemberEntry",
    "DuplicateHealthMonitor",
    "DuplicatePoolEntry",
    "PoolInUseByL7Policy",
    "ImmutableObject",
    "LBPendingStateError",
    "TooManyL7RulesOnL7Policy",
    "ComputeBuildException",
    "ComputeBuildQueueTimeoutException",
    "ComputeDeleteException",
    "ComputeGetException",
    "ComputeStatusException",
    "ComputeGetInterfaceException",
    "IDAlreadyExists",
    "RecordAlreadyExists",
    "NoReadyAmphoraeException",
    "ImageGetException",
    "ComputeWaitTimeoutException",
    "ComputePortInUseException",
    "ComputeUnknownException",
    "InvalidTopology",
    "InvalidL7PolicyAction",
    "InvalidL7PolicyArgs",
    "InvalidURL",
    "InvalidURLPath",
    "InvalidString",
    "InvalidRegex",
    "InvalidL7Rule",
    "ServerGroupObjectCreateException",
    "ServerGroupObjectDeleteException",
    "InvalidAmphoraOperatingSystem",
    "QuotaException",
    "ProjectBusyException",
    "MissingProjectID",
    "MissingAPIProjectID",
    "InvalidSubresource",
    "ValidationException",
    "VIPValidationException",
    "InvalidSortKey",
    "InvalidSortDirection",
    "InvalidMarker",
    "InvalidLimit",
    "MissingVIPSecurityGroup",
    "ProviderNotEnabled",
    "ProviderNotFound",
    "ProviderDriverError",
    "ProviderNotImplementedError",
    "ProviderUnsupportedOptionError",
    "InputFileError",
    "ObjectInUse",
    "ProviderFlavorMismatchError",
    "VolumeDeleteException",
    "VolumeGetException",
    "NetworkServiceError",
    "InvalidIPAddress",
# amphora/driver_exceptions/exceptions.py
    "AmphoraDriverError",
    "NotFoundError",
    "InfoException",
    "MetricsException",
    "UnauthorizedException",
    "StatisticsException",
    "TimeOutException",
    "DeleteFailed",
    "SuspendFailed",
    "EnableFailed",
    "ArchiveException",
    "ProvisioningErrors",
    "ListenerProvisioningError",
    "LoadBalancerProvisoningError",
    "HealthMonitorProvisioningError",
    "NodeProvisioningError",
    "AmpDriverNotImplementedError",
    "AmpConnectionRetry",
# network/base.py
    "NetworkException",
    "PlugVIPException",
    "UnplugVIPException",
    "AllocateVIPException",
    "DeallocateVIPException",
    "PlugNetworkException",
    "UnplugNetworkException",
    "VIPInUseException",
    "PortNotFound",
    "NetworkNotFound",
    "SubnetNotFound",
    "AmphoraNotFound",
    "PluggedVIPNotFound",
    "TimeoutException",
    "QosPolicyNotFound",
    "SecurityGroupNotFound",
    "CreatePortException",
    "AbstractNetworkDriver",
]
